# -*- coding: utf-8 -*-

"""Top-level package for spharpy."""

__author__ = """The pyfar developers"""
__email__ = 'info@pyfar.org'
__version__ = '1.0.0'

from .classes.sh import (
    SphericalHarmonics,
    SphericalHarmonicDefinition,
)
from .classes.coordinates import SamplingSphere
from .classes.audio import (
    SphericalHarmonicSignal,
    SphericalHarmonicTimeData,
    SphericalHarmonicFrequencyData,
)
from . import spherical
from . import samplings
from . import plot
from . import transforms
from . import beamforming
from . import interpolate
from . import spatial
from . import special


__all__ = [
    'SphericalHarmonics',
    'SphericalHarmonicDefinition',
    'SphericalHarmonicSignal',
    'SphericalHarmonicTimeData',
    'SphericalHarmonicFrequencyData',
    'spherical',
    'samplings',
    'plot',
    'transforms',
    'beamforming',
    'interpolate',
    'spatial',
    'special',
    'SamplingSphere',
]
